//
//  main.cpp
//  Blindaje
//
//  Created by Borja Hawk on 28/04/12.
//  Copyright (c) 2012 BHAWK. All rights reserved.
//

#include <iostream>
#include <fstream>
#include <math.h>
#include <vector>
#include <map>
#include <sstream>
#include <mpi.h>
using namespace std;
vector<vector<float> > vcSecciones1;
vector<vector<float> > vcSecciones2;
float round(float r,int n_digit);
int effectfunction(float energia, float *seccionEficaz, bool rayleigh, int tipoEsfera);
float rfunction(float seccionEficaz, float density);
float thetafunction();
float phifunction();
float Efunction(float theta, float E);
float dfunction(float x, float y, float z);
bool cargarFicheroSeccionesEficazes(string composicion, int tipoEsfera);
bool rayleigh = false;
int parserParametros(int argc, char *argv[], int *iteraciones, float *energia, float *radio1, float *radio2, 
                      string * material1, string * material2);
unsigned time_seed(int rank);


int main(int argc, char * argv[])
{
    int rank, size;
     
    //Se inicializa el procesamiento en paralelo
    MPI_Init(&argc, &argv);     
    double timeInit = MPI_Wtime();      //Se toma el tiempo al inicio del programa
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    MPI_Comm_size(MPI_COMM_WORLD,&size);

    int parser = 1;
    int iteraciones = 1000000;          //El número de iteraciones es de 1 millón por defecto
    float energia = 1;                  //La energía inicial es de 1 MeV por defecto
    string composicion1 = "";
    float radio1;
    float density1;
    string composicion2 = "";
    float radio2;
    float density2 = 0;

    //Se obtienen los parámetros introducidos por el usuario, sólo lo hace un nodo
    if(rank == 0)
         parser = parserParametros(argc, argv, &iteraciones, &energia, &radio1, &radio2, &composicion1, &composicion2);
    
    MPI_Bcast(&parser, 1, MPI_INT, 0, MPI_COMM_WORLD);
    if(!parser)                         //En caso de no ser correctos los parámetros se sale del programa
    {
        MPI_Finalize();
        return -1;
    }
    
    //Se envían las composiciones a todos los nodos
    char m1[3], m2[3];
    strcpy(m1, composicion1.c_str());
    MPI_Bcast(&m1, 3, MPI_CHAR, 0, MPI_COMM_WORLD);
    strcpy(m2, composicion2.c_str());
    MPI_Bcast(&m2, 3, MPI_CHAR, 0, MPI_COMM_WORLD);

    //Se inicializa un mapa con los posibles materiales de la esfera y sus densidades
    typedef map<string, float> MAP_MATERIALES;
    MAP_MATERIALES mMateriales;
    mMateriales.insert(MAP_MATERIALES::value_type("Pb",11.34));
    mMateriales.insert(MAP_MATERIALES::value_type("U",19.05));
    mMateriales.insert(MAP_MATERIALES::value_type("UO2",10.97));
    //Se obtiene la densidad del material 1 a usar
    MAP_MATERIALES::iterator it1 = mMateriales.find(m1);
    if(it1 != mMateriales.end())
        density1 = (*it1).second;
    //Se carga el fichero de las secciones eficaces de la composición de la esfera concéntrica
    cargarFicheroSeccionesEficazes(m1, 1);  
    //Se obtiene la densidad del material 2 a usar si se ha introducido alguno
    it1 = mMateriales.find(m2);
    if(it1 != mMateriales.end())
    {
        density2 = (*it1).second;
        cargarFicheroSeccionesEficazes(m2, 2);
    }

    //Se envían los parámetros a todos los nodos
    MPI_Bcast(&iteraciones, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&energia, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&radio1, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
    MPI_Bcast(&radio2, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
    
    //Se reparten el número de fotones a procesar entre los nodos
    int cuentas = iteraciones / size;
    int inicial = 1 + rank*cuentas;
    int final = (rank+1)*cuentas;
    if(rank == size - 1)
        final = iteraciones;
    

    srand(time_seed(rank+1));          //Se inicializa el número aleatorio utilizando el nº de procesador para que los cálculos sean distintos
    int numFotonesPeligrosos = 0;
    float energiasFotones[iteraciones];             //Vector donde se guardará la energía final de cada fotón
    std::fill_n(energiasFotones, iteraciones, 0);   //Se inicializa con ceros
    
    //Se realizan las iteraciones
    for (int k=inicial; k<=final; k++) 
    {
        float E = energia;      //energía introducida por el usuario
        //Se inicializa el punto del fotón, el efecto sufrido y la distancia recorrida
        float x, y, z;
        x=y=z=0;
        int effect = -1;
        float distance = 0;
        //Este bucle se realiza siempre que no haya efecto fotoeléctrico, la distancia recorrida sea menor que el radio y la energía mayor de 1 keV
        do {
            float seccionEficaz = 0;
            float r = 0;
            //Si la distancia es menor del radio 1 se toman los datos de la esfera interna
            if(distance <= radio1)
            {
                effect = effectfunction(E, &seccionEficaz, rayleigh, 1);	//Se obtiene el efecto a partir de la energía
                r = rfunction(seccionEficaz, density1);     //Se halla r en función de la sección eficaz y de la densidad
            }
            else    ////Si la distancia es mayor del radio 1 se toman los datos de la esfera externa
            {
                effect = effectfunction(E, &seccionEficaz, rayleigh, 2);	//Se obtiene el efecto a partir de la energía
                r = rfunction(seccionEficaz, density2);     //Se halla r en función de la sección eficaz y de la densidad
            }
            
            float theta = thetafunction();                  //Se halla aleatoriamente el ángulo theta
            float phi = phifunction();                      //Se halla phi aleatoriamente
            float xold = x; float yold = y; float zold = z;	//Se actualizan las coordenadas anteriores
			float dx = r*sin(theta)*cos(phi);               //Se calcula el nuevo vector recorrido por el fotón
			float dy = r*sin(theta)*sin(phi);
			float dz = r*cos(theta);
			x = xold + dx;                  //Se calculan las nuevas coordenadas respecto al centro de la circunferencia
			y = yold + dy;
			z = zold + dz;
			distance = dfunction(x,y,z);            //Halla la distancia del fotón respecto al centro de la circunferencia
            if(distance > radio2 && effect == 0)    //Si el fotoeléctrico sale fuera de la esfera, en realidad el fotón ha salido
            {
                effect = 1;                 //Se trata como un Compton que no ha perdido energía
                break;
            }
            else if(effect == 1)                    //Si el efecto es Compton cambia la energía
            {
                float Enew = Efunction(theta, E);	//Se halla la energía del fotón resultante a partir de theta y E
                E = Enew;
            }
        } while (distance <= radio2 && effect != 0 && E > 0.001);

        //Si la energía del fotón es la inicial y ha salido de la esfera se incrementa en 1 el número de fotones peligrosos
        if(E == energia && effect != 0)
            numFotonesPeligrosos++;
        //Se calcula la energía del fotón redondeando a 3 decimales si no ha sido efecto fotoeléctrico
        if(effect!=0)
        {
            float Efinal = round(E,3);
            energiasFotones[k-1] = Efinal;
        }
        else
            energiasFotones[k-1] = 0;
    }

    //Se suman el número de fotones peligrosos y los vectores de energías de cada nodo, juntando todos los resultados en el nodo 0
    int numFotonesPeligrososTotal = 0;
    float energiasFotonesTotal[iteraciones];
    MPI_Reduce(&numFotonesPeligrosos, &numFotonesPeligrososTotal, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(&energiasFotones, &energiasFotonesTotal, iteraciones, MPI_FLOAT, MPI_SUM, 0, MPI_COMM_WORLD); 
    
    //Finalmente se crea el archivo con el espectro en el nodo 0
    if (rank==0) 
    {
        //Primero se halla el mapa con el espectro resultante
        typedef map<float, int> MAP_ESPECTRO;
        MAP_ESPECTRO mEspectro;
        for(int i=0; i<iteraciones; i++)
        {
            if(energiasFotonesTotal[i] != 0)
            {
                //Se suma un fotón a la energía saliente dentro del mapa del espectro
                MAP_ESPECTRO::iterator it = mEspectro.find(energiasFotonesTotal[i]);
                if(it != mEspectro.end())
                    (*it).second++;
                else
                    mEspectro.insert(MAP_ESPECTRO::value_type(energiasFotonesTotal[i],1));
            }
        }
        //Después se crea el fichero en el que escribimos el mapa guardado
        ofstream f2;
        stringstream ss;
        ss << "/home/borja/simulaciones/sim_" << composicion1 <<"_radio" << radio1 << composicion2 <<"_radio" << radio2 << "_energia" << energia << "_fotones" << iteraciones << ".txt";
        //ss << "/Users/BorjaHawk/Desktop/Trabajo Master/Algoritmo de blindaje/Simulaciones/sim_" << composicion1 <<"_radio" << radio1 << composicion2 <<"_radio" << radio2 << "_energia" << energia << "_fotones" << iteraciones << ".txt";
        f2.open(ss.str().c_str(), ofstream::out);
        MAP_ESPECTRO::const_iterator it;
        for (it=mEspectro.begin() ; it != mEspectro.end(); it++)
            f2 << (*it).first << " " << (*it).second << endl;
        f2.close();
        
        //Se escriben los resultados de blindaje y tiempo de ejecución por pantalla
        double timeEnd = MPI_Wtime();           //Se guarda el tiempo de finalización
        double timeTotal = timeEnd-timeInit;    //Se halla el tiempo total de ejecución del nodo
        cout << "Tiempo de ejecución de nodo " << rank <<": " << timeTotal << " segundos." << endl;
        float blindaje = round((float)(iteraciones - numFotonesPeligrososTotal)/iteraciones*100, 2);
        cout << "Blindaje (porcentaje de fotones retenidos) = " << blindaje << endl;
        cout << "Archivo de espectro: sim_" << composicion1 <<"_radio" << radio1 << composicion2 <<"_radio" << radio2 << "_energia" << energia << "_fotones" << iteraciones << ".txt" << endl;
    }
    else    //En el resto de nodos solo se escribe el tiempo de ejecución
    {
        double timeEnd = MPI_Wtime();
        double timeTotal = timeEnd-timeInit;
        cout << "Tiempo de ejecución de nodo " << rank <<": " << timeTotal << " segundos." << endl;
    }
    
    MPI_Finalize();     //Se termina la parelelización
    return 1;
}

//Función que hace el redondeo de un número con el número de decimales requerido
float round(float r,int n_digit)
{
    int n=pow(10,n_digit);
    r=((float)((int)(r*n+0.5)))/n;
    return(r);
}

//Función que obtiene el efecto probabilísticamente y además devuelve su sección eficaz a la energía dada
int effectfunction(float energia, float *seccionEficaz, bool rayleigh, int tipoEsfera)
{
    //Dependiendo de la esfera donde esté el fotón se usan las secciones eficaces de una u otra
    vector<vector<float> > vcSecciones;
    if(tipoEsfera == 1)
        vcSecciones = vcSecciones1;
    else if (tipoEsfera == 2)
        vcSecciones = vcSecciones2;
    
    float nr = ((float)rand())/(float)RAND_MAX;
    //se buscan las secciones eficaces correspondientes a la energía del fotón
    for (size_t i=0; i<vcSecciones.size(); i++)
    {
        if(energia < vcSecciones[i][0] && i == 0)  //la energía es más baja que la primera, consideramos absorción
        {
            *seccionEficaz = vcSecciones[i][3];
            return 0;
        }
        else if(energia == vcSecciones[i][0])       //La energía coincide con una de las guardadas
        {
            float sumaSecciones = vcSecciones[i][2] + vcSecciones[i][3];
            if(rayleigh)
                sumaSecciones = sumaSecciones + vcSecciones[i][1];
            float probPE = vcSecciones[i][3] / sumaSecciones;
            float probCO = vcSecciones[i][2] / sumaSecciones;
            if (nr < probPE)                                        //fotoeléctrico
            {
                *seccionEficaz = vcSecciones[i][3];
                return 0;
            }
            else if(nr>=probPE && nr <= probPE + probCO)            //Compton
            {
                *seccionEficaz = vcSecciones[i][2];
                return 1;
            }
            else                                                    //Rayleigh
            {
                *seccionEficaz = vcSecciones[i][1];
                return 2;
            }
        }
        else if(energia > vcSecciones[i][0])
        {
            if(i == vcSecciones.size()-1)   //la energía es más alta que todas, cogemos la última
            {
                float sumaSecciones = vcSecciones[i][2] + vcSecciones[i][3];
                if(rayleigh)
                    sumaSecciones = sumaSecciones + vcSecciones[i][1];
                float probPE = vcSecciones[i][3] / sumaSecciones;
                float probCO = vcSecciones[i][2] / sumaSecciones;
                if (nr < probPE)                                    //fotoeléctrico
                {
                    *seccionEficaz = vcSecciones[i][3];
                    return 0;
                }
                else if(nr>=probPE && nr <= probPE + probCO)        //Compton
                {
                    *seccionEficaz = vcSecciones[i][2];
                    return 1;
                }
                else                                                //Rayleigh
                {
                    *seccionEficaz = vcSecciones[i][1];
                    return 2;
                }
            }
        }
        else if(energia < vcSecciones[i][0])        //La energía es menor que la iteración de la energía
        {
            //Si la energía está entre dos de la lista, se hace el promedio de secciones eficaces
            float promedioPE = vcSecciones[i-1][3];
            float promedioCO = vcSecciones[i-1][2];
            float promedioRA = vcSecciones[i-1][1];
            
            //Hay que comprobar que no haya más energías iguales a la superior
            size_t k = i;
            float seccionAltaPE = vcSecciones[k][3];
            float seccionAltaCO = vcSecciones[k][2];
            float seccionAltaRA = vcSecciones[k][1];
            while(k < vcSecciones.size()-1 && vcSecciones[k][0] == vcSecciones[k+1][0])
            {
                //En caso de haber distintas secciones eficaces con la misma energía se toma la sección más grande
                if (vcSecciones[k+1][3]>seccionAltaPE)
                    seccionAltaPE = vcSecciones[k+1][3];
                if (vcSecciones[k+1][2]>seccionAltaCO)
                    seccionAltaCO = vcSecciones[k+1][2];
                if (vcSecciones[k+1][1]>seccionAltaRA)
                    seccionAltaRA = vcSecciones[k+1][1];
                k++;
            }
            //Se promedian las secciones 
            promedioPE+=seccionAltaPE;
            promedioCO+=seccionAltaCO;
            promedioRA+=seccionAltaRA;
            promedioPE = promedioPE/2;
            promedioCO = promedioCO/2;
            promedioRA = promedioRA/2;
            
            float sumaSecciones = promedioPE + promedioCO;
            if(rayleigh)
                sumaSecciones = sumaSecciones + promedioRA;
            float probPE = promedioPE/sumaSecciones;
            float probCO = promedioCO/sumaSecciones;
            if (nr < probPE)                                        //fotoeléctrico
            {
                *seccionEficaz = promedioPE;
                return 0;
            }
            else if(nr>=probPE && nr <= probPE + probCO)            //Compton
            {
                *seccionEficaz = promedioCO;
                return 1;
            }
            else                                                    //Rayleigh
            {
                *seccionEficaz = promedioRA;
                return 2;
            }
        }
    }
    return -1;
}

//Función que calcula el radio del nuevo vector del fotón
float rfunction(float seccionEficaz, float density)
{
	float mfpath = 1.0 / (seccionEficaz * density);             //fórmula del camino medio libre
	float nr = ((float)rand())/(float)RAND_MAX;					//nº aleatorio entre 0 y 1
	float r = log(nr) * -1 * mfpath;                            //trasladamos el nº aleatorio al intervalo (0, 2*mfpath)
	return r;
}

//Función que calcula theta
float thetafunction()
{
	float nr = ((float)rand())/(float)RAND_MAX;					//nº aleatorio entre 0 y 1
	float theta = nr * M_PI;                                    //se traslada al intervalo (0, pi)
    return theta;
}

//Función que calcula phi
float phifunction()
{
	float nr = ((float)rand())/(float)RAND_MAX;                 //nº aleatorio entre 0 y 1
	float phi = nr * 2 * M_PI;                                  //se traslada al intervalo (0, 2*pi)
    return phi;
}

//Función que calcula la energía nueva después de un efecto Compton
float Efunction(float theta, float E)
{
	float Enew = E / (1 + ((1 - cos(theta)) * E/0.511));		//fórmula de la energía
	return Enew;
}

//Función que calcula la distancia del fotón al centro de la esfera
float dfunction(float x, float y, float z)
{
	float d = sqrt(pow(x,2) + pow(y,2) + pow(z,2));
	return d;
}

//Función usada en parserParametros para obtener los parámetros
char* getCmdOption(char ** begin, char ** end, const std::string & option)
{
    char ** itr = std::find(begin, end, option);
    if (itr != end && ++itr != end)
    {
        return *itr;
    }
    return 0;
}

//Función usada en parserParametros para comprobar la introducción de los parámetros
bool cmdOptionExists(char** begin, char** end, const std::string& option)
{
    return std::find(begin, end, option) != end;
}

//Función usada en parserParametros para comprobar si el parámetro introducido es un número
template <class T>
bool from_string(T& t, const string& s, ios_base& (*f)(ios_base&))
{
    istringstream iss(s);
    return !(iss >> f >> t).fail();
}

//Función que recoge los parámetros introducidos y los guarda en variables
int parserParametros(int argc, char *argv[], int *iteraciones, float *energia, float *radio1, float *radio2, 
                      string * material1, string * material2)
{
    //Se comprueba que al menos se ha introducido el material y el radio de la esfera 1
    if(cmdOptionExists(argv, argv+argc, "-m1") && cmdOptionExists(argv, argv+argc, "-r1"))
    {
        char * charIteraciones = getCmdOption(argv, argv + argc, "-p");     //número de iteraciones
        if (charIteraciones)
        {
            if(!from_string<int>(*iteraciones, charIteraciones, std::dec))
            {
                cout << "El formato de n no es válido." << std::endl;
                return 0;
            }
            else 
                if(iteraciones == 0)
                {
                    cout << "El número de fotones no puede ser 0." << std::endl;
                    return 0;
                }
        }
        char * charEnergia = getCmdOption(argv, argv + argc, "-e");         //energía inicial
        if (charEnergia)
        {
            if(!from_string<float>(*energia, charEnergia, std::dec))
            {
                cout << "El formato de e no es válido." << std::endl;
                return 0;
            }
            else 
            {
                if(energia == 0)
                {
                    cout << "Error - La energía no puede ser 0." << std::endl;
                    return 0;
                }
                else if(*energia > 1)
                {
                    cout << "Error - La energía no puede ser mayor de 1 MeV." << std::endl;
                    return 0;
                }
            }
        }
        char * charMaterial = getCmdOption(argv, argv + argc, "-m1");   //material esfera interna
        if (charMaterial)
        {
            if(strcmp(charMaterial, "Pb") == 0 || strcmp(charMaterial, "U") == 0 || strcmp(charMaterial, "UO2") == 0)
                *material1 = charMaterial;
            else
            {
                cout << "Error - El material de la esfera concéntrica debe ser Pb, U o UO2." << endl;
                return 0;
            }
            
        }
        char * charRadio = getCmdOption(argv, argv + argc, "-r1");      //radio esfera interna
        if (charRadio)
        {
            if(!from_string<float>(*radio1, charRadio, std::dec))
            {
                cout << "El formato de r1 no es válido." << std::endl;
                return 0;
            }
            else 
                if(radio1 == 0)
                {
                    cout << "Error - r1 no puede ser 0." << std::endl;
                    return 0;
                }
        }
        charMaterial = getCmdOption(argv, argv + argc, "-m2");          //material esfera externa
        if (charMaterial)
        {
            if(strcmp(charMaterial, "Pb") == 0 || strcmp(charMaterial, "U") == 0 || strcmp(charMaterial, "UO2") == 0)
            {
                *material2 = charMaterial;
                charRadio = getCmdOption(argv, argv + argc, "-r2");     //radio esfera externa
                if (charRadio)
                {
                    if(!from_string<float>(*radio2, charRadio, std::dec))
                    {
                        cout << "El formato de r2 no es válido." << std::endl;
                        return 0;
                    }
                    else 
                        if(*radio2 <= *radio1)
                        {
                            cout << "Error - r2 debe ser mayor que r1." << std::endl;
                            return 0;
                        }
                }
                else 
                {
                    cout << "Error - Se ha definido m2 pero no r2." << std::endl;
                    return 0;
                }
            }
            else
            {
                cout << "Error - El material de la esfera excéntrica debe ser Pb, U o UO2." << endl;
                return 0;
            }
        }
        else
            *radio2 = *radio1;
        char * charRayleigh = getCmdOption(argv, argv + argc, "-r");    //Con o sin dispersión Rayleigh
        if (charRayleigh)
        {
            if(strcmp(charRayleigh, "s") == 0)
                rayleigh = true;
            else if(strcmp(charRayleigh, "n") != 0)
            {
                cout << "Error - El efecto Rayleigh debe ser s (si) o n (no)." << endl;
                return 0;
            }
            
        }
    }
    else    //Si no se han introducido bien los parámetros salta el texto de ayuda
    {
        cout << "Los parámetros que acepta el programa son los siguientes:" <<endl;
        cout << "-p número de fotones a simular (1 millón si se omite)," << endl;
        cout << "-e energía inicial de los fotones, hasta 1 MeV (1 MeV si se omite)" << endl;
        cout << "-m1 composición de la esfera concéntrica de blindaje (obligatorio), opciones" << endl;
        cout << "    Pb: plomo," << endl << "    U: uranio empobrecido," <<endl << "    UO2: dióxido de uranio;" << endl;
        cout << "-m2 composición de la esfera excéntrica de blindaje (no hay si se omite)," << endl;
        cout << "-r1 radio de la esfera concéntrica de blindaje (obligatorio)." <<endl;
        cout << "-r2 radio de la esfera excéntrica de blindaje (obligatorio si se usa dicha esfera)." <<endl;
        cout << " -r s/n para tener en cuenta o no respectivamente la dispersión de Rayleigh (no si se omite)." << endl;
        return 0;
    }
    return 1;
}

//Función que carga las secciones eficaces en vectores a partir de los ficheros de texto
bool cargarFicheroSeccionesEficazes(string composicion, int tipoEsfera)
{
    vector<vector<float> > vcSecciones;
    string cadenaFichero = "/home/borja/seccioneseficaces/cross-sections-" + composicion + ".dat";
    //string cadenaFichero = "/Users/BorjaHawk/Desktop/Trabajo Master/Algoritmo de blindaje/secciones_eficaces/cross-sections-" + composicion + ".dat";
    
    ifstream fic(cadenaFichero.c_str());
    if (fic == NULL) 
    {
        perror ("Error al abrir el archivo");
        fic.close();
        return false;
    }
    char buffer[100];
    int i = 0;
    
    while(!fic.eof())
    {
        fic.getline(buffer,100);

        {
            char * pch;
            pch = strtok (buffer," ");
            int j = 0;
            vector<float> seccion;
            while (pch != NULL)
            {
                seccion.push_back(atof(pch));
                pch = strtok (NULL, " ");
                j++;
            } 
            vcSecciones.push_back(seccion);
        }
        i++;
    }
    fic.close();
    vcSecciones.pop_back(); //Por razones del strtok se añade un último elemento vacío y hay que quitarlo
    if(tipoEsfera == 1)
        vcSecciones1 = vcSecciones;
    else if (tipoEsfera == 2)
        vcSecciones2 = vcSecciones;
    
    return true;
}

//Función que inicia la semilla de los números pseudoaleatorios a partir del número de nodo
unsigned time_seed(int rank)
{
    time_t now =time(0);
    unsigned char *p = (unsigned char *)&now;
    
    unsigned seed=rank;
    size_t i;
    
    for (i=0;i<sizeof now;i++)
        seed = seed * (UCHAR_MAX+2U) + p[i];
    return seed;
}

