//
//  main.cpp
//  NaI
//
//  Created by Borja Hawk on 10/04/12.
//  Copyright (c) 2012 BHAWK. All rights reserved.
//

#include <iostream>
#include <fstream>
#include <math.h>
#include <vector>
#include <map>
using namespace std;
vector<vector<float> > vcSecciones;
float round(float r,int n_digit);
int effectfunction(float energia, float *seccionEficaz, bool rayleigh);
float rfunction(float seccionEficaz, float density);
float thetafunction();
float phifunction();
float Efunction(float theta, float E);
float dfunction(float x, float y, float z);
bool rayleigh = false;  //Variable que tiene en cuenta o no la dispersión Rayleigh


int main(int argc, const char * argv[])
{
    //Se carga el fichero de las secciones eficaces del NaI
    ifstream fic("/Users/BorjaHawk/Downloads/cross-sections-nai.dat");
    if (fic == NULL) 
        perror ("Error al abrir el archivo");
    char buffer[100];
    int i = 0;
    
    while(!fic.eof())
    {
        fic.getline(buffer,100);
        
        if(i>2) //No se tienen en cuenta las 2 primeras líneas
        {
            char * pch;
            pch = strtok (buffer," ");
            int j = 0;
            vector<float> seccion;
            while (pch != NULL)
            {
                seccion.push_back(atof(pch));                
                pch = strtok (NULL, " ");
                j++;
            } 
            vcSecciones.push_back(seccion);
        }
        i++;
    }
    fic.close();
    vcSecciones.pop_back(); //Por razones del strtok se añade un último elemento vacío y hay que quitarlo
    
     
    //Se realizan las iteraciones
    int iteraciones = 1000000;
    int radio = 8;
    typedef map<float, int> MAP_ESPECTRO;   //Se crea el mapa que almacenará el espectro energético
    MAP_ESPECTRO mEspectro;
    float density = 3.67;
    int numPE = 0;
    srand(time(NULL));          //Se inicializa el número aleatorio

    for (int k=0; k<iteraciones; k++) 
    {
        float E = 0.662;      //energía del fotón de Cs137
        float x, y, z;
        x=y=z=0;
        int effect = -1;
        float distance = 0;
        //Este bucle se realiza siempre que no haya efecto fotoeléctrico y la distancia recorrida sea menor que el radio
        do {
            float seccionEficaz = 0;
            effect = effectfunction(E, &seccionEficaz, rayleigh);	//Se obtiene el efecto a partir de la energía del fotón  
                                                            //y la sección eficaz. Se indica si se quiere Rayleigh
            float r = rfunction(seccionEficaz, density);	//Se halla r en función de la sección eficaz y de la densidad
            float theta = thetafunction();                  //Se halla aleatoriamente el ángulo theta
            float phi = phifunction();                      //Se halla phi aleatoriamente
            float xold = x; float yold = y; float zold = z;	//Se actualizan las coordenadas anteriores
			float dx = r*sin(theta)*cos(phi);               //Se calcula el nuevo vector recorrido por el fotón
			float dy = r*sin(theta)*sin(phi);
			float dz = r*cos(theta);
			x = xold + dx;                  //Se calculan las nuevas coordenadas respecto al centro de la circunferencia
			y = yold + dy;
			z = zold + dz;
			distance = dfunction(x,y,z);	//Halla la distancia del fotón respecto al centro de la circunferencia
            if(distance > radio && effect == 0) //Si el fotoeléctrico sale fuera de la esfera, en realidad el fotón ha salido
            {
                effect = 5;                 //Se trata como un fotón que no ha salido y no ha perdido energía
                break;
            }
            else if(effect == 1)                //Si el efecto es Compton cambia la energía
            {
                float Enew = Efunction(theta, E);	//Se halla la energía del fotón resultante a partir de theta y E
                E = Enew;
            }
        } while (distance <= radio && effect != 0);
        
        if(effect==0)   //Si se ha producido una absorción se incrementa el número de absorciones
            numPE++;
        else if(effect == 1)                        //efecto Compton -> Hay que obtener el espectro
        {
            //Se calcula la energía depositada redondeando a 3 decimales
            float Edepositada = round(0.662 - E,3);
            //Se suma un fotón a la energía depositada dentro del mapa del espectro
            MAP_ESPECTRO::iterator it = mEspectro.find(Edepositada);
            if(it != mEspectro.end())
                (*it).second++;
            else
                mEspectro.insert(MAP_ESPECTRO::value_type(Edepositada,1));
        }

    }
    //Se copia en el mapa la energía de fotoabsorción
    mEspectro.insert(MAP_ESPECTRO::value_type(0.661,0));    //Se introduce este punto para que salga el fotopico como una delta
    mEspectro.insert(MAP_ESPECTRO::value_type(0.662,numPE));
    //Finalmente se crea el archivo con el espectro
    ofstream f2;
    f2.open("/Users/BorjaHawk/Documents/ficheroEspectro.txt", ofstream::out);
    MAP_ESPECTRO::const_iterator it;
    for (it=mEspectro.begin() ; it != mEspectro.end(); it++)
        f2 << (*it).first << " " << (*it).second << endl;
    for(float i = 0.663; i<= 0.7; i+=0.001)
        f2 << i << " " << 0 << endl;
    f2.close();
        
    return 0;
}

//Función que hace el redondeo de un número con el número de decimales requerido
float round(float r,int n_digit)
{
    int n=pow(10,n_digit);
    r=((float)((int)(r*n+0.5)))/n;
    return(r);
}

//Función que obtiene el efecto probabilísticamente y además devuelve su sección eficaz a la energía dada
int effectfunction(float energia, float *seccionEficaz, bool rayleigh)
{
    float nr = ((float)rand())/(float)RAND_MAX;
    //se buscan las secciones eficaces correspondientes a la energía del fotón
    for (size_t i=0; i<vcSecciones.size(); i++)
    {
        if(energia < vcSecciones[i][0] && i == 0)  //la energía es más baja que la primera, consideramos absorción
        {
            *seccionEficaz = vcSecciones[i][3];
            return 0;
        }
        else if(energia == vcSecciones[i][0])       //La energía coincide con una de las guardadas
        {
            float sumaSecciones = vcSecciones[i][2] + vcSecciones[i][3];
            if(rayleigh)
                sumaSecciones = sumaSecciones + vcSecciones[i][1];
            float probPE = vcSecciones[i][3] / sumaSecciones;
            float probCO = vcSecciones[i][2] / sumaSecciones;
            if (nr < probPE)                                        //fotoeléctrico
            {
                *seccionEficaz = vcSecciones[i][3];
                return 0;
            }
            else if(nr>=probPE && nr <= probPE + probCO)            //Compton
            {
                *seccionEficaz = vcSecciones[i][2];
                return 1;
            }
            else                                                    //Rayleigh
            {
                *seccionEficaz = vcSecciones[i][1];
                return 2;
            }
        }
        else if(energia > vcSecciones[i][0])
        {
            if(i == vcSecciones.size()-1)   //la energía es más alta que todas, cogemos la última
            {
                float sumaSecciones = vcSecciones[i][2] + vcSecciones[i][3];
                if(rayleigh)
                    sumaSecciones = sumaSecciones + vcSecciones[i][1];
                float probPE = vcSecciones[i][3] / sumaSecciones;
                float probCO = vcSecciones[i][2] / sumaSecciones;
                if (nr < probPE)                                    //fotoeléctrico
                {
                    *seccionEficaz = vcSecciones[i][3];
                    return 0;
                }
                else if(nr>=probPE && nr <= probPE + probCO)        //Compton
                {
                    *seccionEficaz = vcSecciones[i][2];
                    return 1;
                }
                else                                                //Rayleigh
                {
                    *seccionEficaz = vcSecciones[i][1];
                    return 2;
                }
            }
        }
        else if(energia < vcSecciones[i][0])    //La energía es menor que la iteración de la energía
        {
            //Si la energía está entre dos de la lista, se hace el promedio de secciones eficaces
            float promedioPE = vcSecciones[i-1][3];
            float promedioCO = vcSecciones[i-1][2];
            float promedioRA = vcSecciones[i-1][1];

            //Hay que comprobar que no haya más energías iguales a la superior
            size_t k = i;
            float seccionAltaPE = vcSecciones[k][3];
            float seccionAltaCO = vcSecciones[k][2];
            float seccionAltaRA = vcSecciones[k][1];
            while(k < vcSecciones.size()-1 && vcSecciones[k][0] == vcSecciones[k+1][0])
            {
                //En caso de haber distintas secciones eficaces con la misma energía se toma la sección más grande
                if (vcSecciones[k+1][3]>seccionAltaPE)
                    seccionAltaPE = vcSecciones[k+1][3];
                if (vcSecciones[k+1][2]>seccionAltaCO)
                    seccionAltaCO = vcSecciones[k+1][2];
                if (vcSecciones[k+1][1]>seccionAltaRA)
                    seccionAltaRA = vcSecciones[k+1][1];
                k++;
            }
            //Se promedian las secciones 
            promedioPE+=seccionAltaPE;
            promedioCO+=seccionAltaCO;
            promedioRA+=seccionAltaRA;
            promedioPE = promedioPE/2;
            promedioCO = promedioCO/2;
            promedioRA = promedioRA/2;
            
            float sumaSecciones = promedioPE + promedioCO;
            if(rayleigh)
                sumaSecciones = sumaSecciones + promedioRA;
            float probPE = promedioPE/sumaSecciones;
            float probCO = promedioCO/sumaSecciones;
            if (nr < probPE)                                        //fotoeléctrico
            {
                *seccionEficaz = promedioPE;
                return 0;
            }
            else if(nr>=probPE && nr <= probPE + probCO)            //Compton
            {
                *seccionEficaz = promedioCO;
                return 1;
            }
            else                                                    //Rayleigh
            {
                *seccionEficaz = promedioRA;
                return 2;
            }
        }
    }
    return -1;
}

//Función que calcula el radio del nuevo vector del fotón
float rfunction(float seccionEficaz, float density)
{
	float mfpath = 1.0 / (seccionEficaz * density);             //fórmula del camino medio libre
	float nr = ((float)rand())/(float)RAND_MAX;					//nº aleatorio entre 0 y 1
	float r = -1 * log(nr) * mfpath;                            //trasladamos el nº aleatorio usando logaritmo
	return r;
}

//Función que calcula theta
float thetafunction()
{
	float nr = ((float)rand())/(float)RAND_MAX;					//nº aleatorio entre 0 y 1
	float theta = nr * M_PI;                                    //se traslada al intervalo (0, pi)
    return theta;
}

//Función que calcula phi
float phifunction()
{
	float nr = ((float)rand())/(float)RAND_MAX;                 //nº aleatorio entre 0 y 1
	float phi = nr * 2 * M_PI;                                  //se traslada al intervalo (0, 2*pi)
    return phi;
}

//Función que calcula la energía nueva después de un efecto Compton
float Efunction(float theta, float E)
{
	float Enew = E / (1 + ((1 - cos(theta)) * E/0.511));		//fórmula de la energía
	return Enew;
}

//Función que calcula la distancia del fotón al centro de la esfera
float dfunction(float x, float y, float z)
{
	float d = sqrt(pow(x,2) + pow(y,2) + pow(z,2));
	return d;
}



